#include <fstream>

#include "ArhWriter.h"
#include "Mesh.h"
#include "MeshStructure.h"
#include "Traverse.h"
#include "DOFVector.h"

namespace AMDiS {
  
  using namespace std;


  void ArhWriter::write(string filename, Mesh *mesh)
  {
    std::vector<DOFVector<double>*> vecs(0);
    ArhWriter::write(filename, mesh, vecs);
  }


  void ArhWriter::write(std::string filename, Mesh *mesh, DOFVector<double>* vec)
  {
    std::vector<DOFVector<double>*> vecs(1);
    vecs[0] = vec;

    ArhWriter::write(filename, mesh, vecs);
  }


  void ArhWriter::write(std::string filename, Mesh *mesh, 
			std::vector<DOFVector<double>*> vecs)
  {
    FUNCNAME("ArhWriter::write()");

    ofstream file;
    file.open(filename.c_str(), ios::out | ios::binary | ios::trunc);

    string typeId = "arhP";
    file.write(typeId.c_str(), 4);

    uint32_t nMacroElements = 0;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, 0, Mesh::CALL_EL_LEVEL);
    while (elInfo) {
      nMacroElements++;
      elInfo = stack.traverseNext(elInfo);
    }

    file.write(reinterpret_cast<char*>(&nMacroElements), 4);
    
    uint32_t nValueVectors = vecs.size();
    file.write(reinterpret_cast<char*>(&nValueVectors), 4);

    uint32_t nAllValues = 
      (vecs.size() > 0 ? vecs[0]->getFeSpace()->getAdmin()->getUsedDOFs() : 0);
    file.write(reinterpret_cast<char*>(&nAllValues), 4);

    MeshStructure elementStructure;
    std::vector<std::vector<double> > values(vecs.size());
    int32_t macroElIndex = -1;

    elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
    while (elInfo) {
      if (elInfo->getLevel() == 0) {
	if (macroElIndex != -1) {
	  elementStructure.commit();
	  writeMacroElement(file, elementStructure, values, macroElIndex);
	}

	elementStructure.clear();
	macroElIndex = elInfo->getElement()->getIndex();
	for (unsigned int i = 0; i < vecs.size(); i++) {
	  values[i].clear();

	  for (int j = 0; j <= mesh->getDim(); j++)
	    values[i].push_back((*vecs[i])[elInfo->getElement()->getDOF(j, 0)]);
	}
      }

      elementStructure.insertElement(elInfo->getElement()->isLeaf());
      
      if (!elInfo->getElement()->isLeaf()) {
	for (unsigned int i = 0; i < vecs.size(); i++)
	  values[i].push_back((*vecs[i])[elInfo->getElement()->getChild(0)->getDOF(mesh->getDim(), 0)]);
      }

      elInfo = stack.traverseNext(elInfo);
    }

    // And write the last macro element to file.
    TEST_EXIT_DBG(macroElIndex != -1)("Should not happen!\n");
    elementStructure.commit();
    writeMacroElement(file, elementStructure, values, macroElIndex);
    
    file.close();

    MSG("ARH file written to: %s\n", filename.c_str());
  }

  
  void ArhWriter::writeMacroElement(std::ofstream &file, 
				    MeshStructure &code,
				    std::vector<std::vector<double> >& values,
				    int32_t elIndex)
  {
    file.write(reinterpret_cast<char*>(&elIndex), 4); 
    
    uint32_t nStructureCodes = code.getCode().size();
    file.write(reinterpret_cast<char*>(&nStructureCodes), 4);
    
    uint32_t codeSize = code.getNumElements();
    file.write(reinterpret_cast<char*>(&codeSize), 4);

    file.write(reinterpret_cast<char*>(&(const_cast<vector<uint64_t>&>(code.getCode())[0])), 
	       8 * nStructureCodes);

    uint32_t nValuesPerVector = (values.size() > 0 ? values[0].size() : 0);
    file.write(reinterpret_cast<char*>(&nValuesPerVector), 4);

    for (unsigned int i = 0; i < values.size(); i++)
      file.write(reinterpret_cast<char*>(&(values[i][0])), 8 * nValuesPerVector);
  }

}
